/*global
  xmat4:false
  get_time:false
  Position:true
  Motion:true
*/


function clamp(x, a, b) {
    "use strict";
    if (x < a) { x = a; }
    if (x > b) { x = b; }
    return x;
}

function abs2(x, y) {
    "use strict";
    return Math.sqrt(x*x + y*y);
}

function abs3(x, y, z) {
    "use strict";
    return Math.sqrt(x*x + y*y + z*z);
}

//=====================================================
// rotations
//=====================================================

/**
   Sets a to the rotation matrix specified by angle, axis.
   Assumes axis = [x, y, 0].
   Assumes axis is normalized.
*/
function rotationR3(a, angle, axis)
{
    "use strict";
    // assumes x^2+y^2+z^2 = 1

    var u = Math.cos(angle), v = Math.sin(angle);
    var u1 = 1.0 - u;

    var x = axis[0], y = axis[1];

    // normalize axix
    var q = Math.sqrt(x*x+y*y);
    x /= q;
    y /= q;


    var xx = x*x, xy = x*y, yy = y*y;

    // first row
    a[0] = xx + yy*u;
    a[4] = xy*u1;
    a[8] = y*v;
    a[12] = 0.0;
    
    // second row
    a[1] = xy*u1;
    a[5] = yy + xx*u;
    a[9] = -x*v;
    a[13] = 0.0;

    // third row
    a[2] = -y*v;
    a[6] = x*v;
    a[10] = u;
    a[14] = 0.0;

    // fourth row
    a[3] = 0.0;
    a[7] = 0.0;
    a[11] = 0.0;
    a[15] = 1.0;
}


/**
   Sets a to the rotation matrix in S3 specified by angle, axis.
   Assumes axis is normalized.
   Assumes axis is [x, y, 0].

   Computes MatrixExp[angle A], where
   A = {{0, 0, 0, -y}, {0, 0, 0, x}, {0, 0, 0, 0}, {y, -x, 0, 0}}
*/
function rotationS3(a, angle, axis) {
    "use strict";
    // assumes x^2+y^2 = 1

    var u = Math.cos(angle), v = Math.sin(angle);
    var u1 = 1.0 - u;

    var x = axis[0], y = axis[1];
    var yy = y*y, yx = y*x, xx = x*x;

    // first row
    a[0] = xx + yy*u;
    a[4] = yx*u1;
    a[8] = 0.0;
    a[12] = y*v;

    // second row
    a[1] = yx*u1;
    a[5] = yy + xx*u;
    a[9] = 0.0;
    a[13] = -x*v;

    // third row
    a[2] = 0.0;
    a[6] = 0.0;
    a[10] = 1.0;
    a[14] = 0.0;

    // fourth row
    a[3] = -y*v;
    a[7] = x*v;
    a[11] = 0.0;
    a[15] = u;
}

function rotationZ(a, angle)
{
    "use strict";
    var u = Math.cos(angle), v = Math.sin(angle);

    // first row
    a[0] = u;
    a[4] = -v;
    a[8] = 0.0;
    a[12] = 0.0;
    
    // second row
    a[1] = v;
    a[5] = u;
    a[9] = 0.0;
    a[13] = 0.0;

    // third row
    a[2] = 0.0;
    a[6] = 0.0;
    a[10] = 1.0;
    a[14] = 0.0;

    // fourth row
    a[3] = 0.0;
    a[7] = 0.0;
    a[11] = 0.0;
    a[15] = 1.0;
}



//=====================================================
// Position
//=====================================================


Position = function() {
    "use strict";
    this.obj_transform = xmat4.create();
    this.cam_transform = xmat4.create();

    this.tmp = xmat4.create();
};
Position.prototype.clear = function() {
    "use strict";
    xmat4.identity(this.obj_transform);
    xmat4.identity(this.cam_transform);
};

/*
  Position.prototype.move_camera = function(angle, translation) {

  this.cam_transform[12] += translation[0];
  this.cam_transform[13] += translation[1];
  this.cam_transform[14] += translation[2];

  if (angle != 0.0)
  {
  // cam_transform = cam_transform * rotation
  rotationZ(this.tmp, angle);
  xmat4.multiply(this.cam_transform, this.cam_transform, this.tmp);
  }
  };
*/

//=====================================================
// Motion
//=====================================================

/*
  Motion represents an incremental motion of one of four types:

  mode == 0: off
  mode == 1: rotationR3               scalar=angle, vector=axis
  mode == 2: rotationZ,translation:   scalar=angle, vector=translation
  mode == 3: rotationS3               scalar=angle, vector=axis
*/

var OFF = 0;
var ROTATE_R3 = 1;
var MOVE_Z = 2;
var ROTATE_S3 = 3;

Motion = function() {
    "use strict";
    this.mode = OFF;
    this.scalar = 0.0;
    this.vector = [0.0, 0.0, 0.0];

    this.zoom_factor = 1.0;
    this.translation_vector = [0.0, 0.0];
    this.rotation_angle = 0.0;

    this.tmp = xmat4.create();
};
Motion.prototype.clear = function() {
    "use strict";
    this.mode = OFF;
    this.scalar = 0.0;
    this.vector = [0.0, 0.0, 0.0];
};
Motion.prototype.set = function(mode, scalar, vector) {
    "use strict";
    this.mode = mode;

    if (mode == MOVE_Z) {

        /*
        // !!! ROT
        this.scalar = scalar;
        */
        // !!! TRANS
        //this.vector = vector;
        this.translation_vector = [ vector[0], vector[1] ];
        // !!! ZOOM
        this.zoom_factor = vector[2];
    }
    else {
        this.scalar = scalar;
        this.vector = vector;
    }
};

Motion.prototype.please_translate = function(dx, dy) {
    this.mode = MOVE_Z;
    this.translation_vector = [dx, dy];
};

Motion.prototype.please_zoom = function(dz) {
    this.mode = MOVE_Z;
    this.zoom_factor = dz;
};

Motion.prototype.please_rotate = function(dangle) {
    this.mode = MOVE_Z;
    this.rotation_angle = dangle;
};

Motion.prototype.magnitude = function() {
    "use strict";
    switch (this.mode) {
    case OFF:
        return 0.0;
    case ROTATE_R3:
        return this.scalar;
    case MOVE_Z:
        //return Math.max( this.scalar, abs3(this.vector[0], this.vector[1], this.vector[2]) );
        return Math.abs(this.zoom_factor);
    case ROTATE_S3:
        return this.scalar;
    }
};

Motion.prototype.decay = function() {
    "use strict";
    switch (this.mode) {
    case OFF:
        break;
    case ROTATE_R3:
        this.scalar *= 0.99;
        break;
    case MOVE_Z:
        // ROT
        this.rotation_angle *= 0.98;

        // TRANS!!!
        var s = 0.97;
        this.translation_vector[0] *= s;
        this.translation_vector[1] *= s;

        // ZOOM!!!
        var t = 0.02;
        this.zoom_factor = (1.0-t)*this.zoom_factor + (t)*1.0;
        break;
    case ROTATE_S3:
        this.scalar *= 0.99;
        break;
    }
};

/*
  Updates position by an incremental motion.
*/
Motion.prototype.increment = function(position) {
    "use strict";

    switch (this.mode) {
    case OFF:
        break;
    case ROTATE_R3:
        // transform = rotation * transform
        rotationR3(this.tmp, this.scalar, this.vector);
        xmat4.multiply(position.obj_transform, this.tmp, position.obj_transform );
        break;
    case MOVE_Z:


        // !!! TRANS
        xmat4.identity(this.tmp);
        this.tmp[12] = this.translation_vector[0];
        this.tmp[13] = this.translation_vector[1];
        xmat4.multiply(position.cam_transform, this.tmp, position.cam_transform);


        // translation cutoff
        var transx_cutoff0 = -3;
        var transx_cutoff1 = 3;
        var transy_cutoff0 = -1.5;
        var transy_cutoff1 = 1.5;
        position.cam_transform[12] = clamp( position.cam_transform[12], transx_cutoff0, transx_cutoff1);
        position.cam_transform[13] = clamp( position.cam_transform[13], transy_cutoff0, transy_cutoff1);


        // !!! ZOOM
        xmat4.identity(this.tmp);
        this.tmp[0] = this.zoom_factor;
        this.tmp[5] = this.zoom_factor;
        this.tmp[10] = this.zoom_factor;
        xmat4.multiply(position.cam_transform, this.tmp, position.cam_transform);

        // cutoff
        var zoom_cutoff0 = 0.2;
        var zoom_cutoff1 = 1.0/0.2;
        if (position.cam_transform[0] < zoom_cutoff0)
        {
            position.cam_transform[0] = zoom_cutoff0;
            position.cam_transform[5] = zoom_cutoff0;
            position.cam_transform[10] = zoom_cutoff0;
        }
        if (position.cam_transform[0] > zoom_cutoff1)
        {
            position.cam_transform[0] = zoom_cutoff1;
            position.cam_transform[5] = zoom_cutoff1;
            position.cam_transform[10] = zoom_cutoff1;
        }


        // ROT
        // transform = Zrotation * transform
        if (this.rotation_angle !== 0.0) {
            rotationZ(this.tmp, this.rotation_angle);
            xmat4.multiply(position.obj_transform, this.tmp, position.obj_transform);
        }

        // =========== JUNK ================
        /*
        // old translation
        position.cam_transform[12] += this.vector[0];
        position.cam_transform[13] += this.vector[1];
        position.cam_transform[14] += this.vector[2];

        // transform = Zrotation * transform
        if (this.scalar !== 0.0) {
            rotationZ(this.tmp, this.scalar);
            xmat4.multiply(position.obj_transform, this.tmp, position.obj_transform);
        }
        */

        break;
    case ROTATE_S3:
        // S3 rotation
        // transform = rotation * transform
        rotationS3(this.tmp, this.scalar, this.vector);
        xmat4.multiply(position.obj_transform, this.tmp, position.obj_transform );
        break;
    }

    this.decay();
};

//=====================================================
// Animator
//=====================================================

XLAB.Animator = function() {
    "use strict";
    this.max_angle = 200.0;

    this.rotationR3_speed = 0.0005;
    this.rotationS3_speed = 0.0004;
    this.translation_speed = 0.005;
    this.zoom_speed = 0.01;

    this.time_cutoff = 200;

    this.position = new Position();
    this.motion = new Motion();

    this.animating = false;
    this.time = 0;

    /*
      this.mode = 0;

      // object transform
      this.axis = [1.0, 0.0, 0.0];
      this.angle = 0.0;

      // camera transform
      this.cam_angle = 0.0;
      this.cam_translation = [0.0, 0.0, 0.0];

      this._rotate = rotationR3;
      this.speed = this.default_speed[0];

      this.position = position;

      this.tmp = xmat4.create();
    */
};

XLAB.Animator.prototype.clear = function() {
    "use strict";
    this.time = 0;
    this.animating = false;
    this.position.clear();
    this.motion.clear();
};

XLAB.Animator.prototype.get_transforms = function() {
    "use strict";
    if (this.animating) {
        var t = get_time();
        var dt = t - this.time;
        this.motion.increment(this.position);
        this.time = t;
    }
    return [this.position.obj_transform, this.position.cam_transform];
};



/*
  XLAB.Animator.prototype.set_s3 = function(state) {
  if (state) {
  // S^3 mode
  this._rotate = rotationS3;
  this.speed = this.default_speed[0];
  }
  else {
  // R^3 mode
  this._rotate = rotationR3;
  this.speed = this.default_speed[1];
  }
  };
*/

XLAB.Animator.prototype._start = function(mode) {
    "use strict";
    /*
      if (this.animating)
      {
      // update the transform
      xmat4.copy(this.position.obj_transform, this.get_transform());
      this.animating = false;
      }
      this.angle = 0.0;
      this.axis = [1.0, 0.0, 0.0];
      this.cam_angle = 0.0;
      this.cam_translation = [0.0, 0.0, 0.0];
    */
    this.animating = false;
    this.time = get_time();
    this.motion.set(mode, 0.0, [1.0, 0.0, 0.0]);
};

XLAB.Animator.prototype._move = function(mode, scalar, vector) {
    this.animating = false;
    this.time = get_time();
    this.motion.set(mode, scalar, vector);
    this.motion.increment(this.position);
};
XLAB.Animator.prototype.please_translate = function(dx, dy) {
    this.animating = false;
    this.time = get_time();
    this.motion.please_translate(dx, dy);
    this.motion.increment(this.position);
};
XLAB.Animator.prototype.please_zoom = function(dz) {
    this.animating = false;
    this.time = get_time();
    this.motion.please_zoom(dz);
    this.motion.increment(this.position);
};
XLAB.Animator.prototype.please_rotate = function(dangle) {
    this.animating = false;
    this.time = get_time();
    this.motion.please_rotate(dangle);
    this.motion.increment(this.position);
};

XLAB.Animator.prototype._end = function() {
    "use strict";

    // start animation if motion is large and time interval is short
    this.animating = false;

    var t = get_time();
    var dt = t - this.time;

    this.animating =
        //this.motion.mode == 1 &&
        //(this.motion.magnitude() > 0.002)
        //&&
        (dt < this.time_cutoff);
    //this.animating = true;

/*
    if (this.mode === 0)
    {
        if (this.angle !== 0.0)
        {
            if (dt < this.time_cutoff) {
                console.log("start animation 0");
                this.animating = true;
            }
        }
    }
    else if (this.mode == 1)
    {
        if (true) // should check cutoff
        {
            if (dt < this.time_cutoff) {
                console.log("start animation 1 angle=" +
                            this.cam_angle + " translation=" +
                            " " + this.cam_translation[0] +
                            " " + this.cam_translation[1] +
                            " " + this.cam_translation[2] );
                this.animating = true;
            }
        }
    }
    this.time = t;
*/
};

// ----------------------------
// rotate (R3 or S3)

XLAB.Animator.prototype._rotate_start = function(mode, p) {
    "use strict";
    this._start(mode);
};

// move from p=[x0,y0] to q=[x1,y1]
XLAB.Animator.prototype._rotate_move = function(mode, p, q) {
    "use strict";
    var dx = q[0] - p[0];
    var dy = q[1] - p[1];
    var angle = abs2(dx, dy);
    // check for division by zero
    if (angle === 0.0) {
        this._move(mode, 0.0, [1.0, 0.0, 0.0]);
    }
    else {
        var axis = [dy/angle, dx/angle, 0];
        angle = clamp(angle, 0.0, this.max_angle);
        var speed = mode==ROTATE_S3 ? this.rotationS3_speed : this.rotationR3_speed;
        angle *= speed;
        this._move(mode, angle, axis);
    }
};

XLAB.Animator.prototype._rotate_end = function(mode) {
    "use strict";
    this._end();
};


// ----------------------------
// rotateR3

XLAB.Animator.prototype.rotateR3_start = function(p) {
    "use strict";
    this._rotate_start(ROTATE_R3, p);
};

// move from p=[x0,y0] to q=[x1,y1]
XLAB.Animator.prototype.rotateR3_move = function(p, q) {
    "use strict";
    this._rotate_move(ROTATE_R3, p, q);
};

XLAB.Animator.prototype.rotateR3_end = function() {
    "use strict";
    this._rotate_end(ROTATE_R3);
};

// ----------------------------
// rotateS3

XLAB.Animator.prototype.rotateS3_start = function(p) {
    "use strict";
    this._rotate_start(ROTATE_S3, p);
};

// move from p=[x0,y0] to q=[x1,y1]
XLAB.Animator.prototype.rotateS3_move = function(p, q) {
    "use strict";
    this._rotate_move(ROTATE_S3, p, q);
};

XLAB.Animator.prototype.rotateS3_end = function() {
    "use strict";
    this._rotate_end(ROTATE_S3);
};

// --------------------
// translate

XLAB.Animator.prototype.translateXY_start = function(p) {
    "use strict";
    this._start(MOVE_Z);
};

// move from p=[x0,y0] to q=[x1,y1]
XLAB.Animator.prototype.translateXY_move = function(p, q) {
    "use strict";
    var s = this.translation_speed;
    this._move(MOVE_Z, 0.0, [s*(q[0]-p[0]), -s*(q[1]-p[1]), 0.0]);
};

XLAB.Animator.prototype.translateXY_end = function() {
    "use strict";
    this._end();
};



XLAB.Animator.prototype.translateXY = function(dv) {
    "use strict";
    //this._move(MOVE_Z, 0.0, [dv[0], dv[1], 0.0]);
    this.please_translate(dv[0], dv[1]);
};
XLAB.Animator.prototype.translateZ = function(dz) {
    "use strict";
    var factor = 0.1;
    var x = Math.exp(factor*(-dz));

    //this._move(MOVE_Z, 0.0, [0.0, 0.0, x]);
    this.please_zoom(x);
};



XLAB.Animator.prototype.rotateZ = function(dangle) {
    "use strict";
    //this._move(MOVE_Z, dangle, [0.0, 0.0, 0.0]);
    this.please_rotate(dangle);
};


// --------------------
// 2finger touch

XLAB.Animator.prototype.twofinger_start = function(a, b) {
    "use strict";
    this._start(ROTATE_S3);
};

XLAB.Animator.prototype.twofinger_move = function(a0, b0, a1, b1) {
    "use strict";
    // previous length
    var dx = a0[0]-b0[0];
    var dy = a0[1]-b0[1];
    var d0 = abs2(dx, dy);

    // current length
    dx = a1[0]-b1[0];
    dy = a1[1]-b1[1];
    var d1 = abs2(dx, dy);

    /*
    // var speed = (this.cam_translation[2] + 100)/300;
    var speed = (this.position.cam_transform[14] + 30)/300.0;
    if (speed > 1.0) { speed = 1.0; }
    if (speed < 0.01) { speed = 0.01; }
    var zoom = speed*(d0-d1);
    */
    var zoom = this.zoom_speed*(d0-d1);

    //----------------
    // rotation angle (note atan2(y,x) is backwards order)
    var angle0 = Math.atan2(b0[1]-a0[1], b0[0]-a0[0]);
    var angle1 = Math.atan2(b1[1]-a1[1], b1[0]-a1[0]);
    var angle = -(angle1 - angle0);

    //----------------
    // translation
    // compute average position of the two fingers
    var x0 = 0.5*(a0[0] + b0[0]);
    var y0 = 0.5*(a0[1] + b0[1]);
    var x1 = 0.5*(a1[0] + b1[0]);
    var y1 = 0.5*(a1[1] + b1[1]);

    zoom = clamp( zoom, -20.0, 200.0 );

    var s = this.translation_speed;
    var t = [s*(x1-x0), -s*(y1-y0), -zoom];

    this._move(MOVE_Z, angle, t);
};

XLAB.Animator.prototype.twofinger_end = function() {
    "use strict";
    this._end();
};
